#!/usr/bin/perl -w
# minorweb.pl - This is minorweb, a web interface to minordomo
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
use CGI;
use Socket;
use POSIX;
use strict;

# You may wish to edit this if you are using a different config file
my $config = "/etc/minordomo.conf";

# The version of minordomo
my $version = "minordomo 0.7.6.2";

# Set up default values
my $smtp_server = 'localhost';
my $directory = '/var/lib/minordomo';
my $admin = 'root@localhost';
my $domain = 'localdomain.com';
my @archive = ();
my @list_cmd = ();
my @open_lists = ();
my $subject_pre = '';
my $url = '';
my $message = '';
my $template = '';

# Read the config file
do_conf($config);

$url = 'http://www.ndn.net/perl/minorweb.pl';
my $minor_url = $url;

my $action = CGI::param('A') || 'LIST';
my $list   = CGI::param('L');
my $year   = CGI::param('Y') || 'NONE';
my $month  = CGI::param('M') || 'NONE';
my $daydir = CGI::param('D') || 'NONE';
my $id     = CGI::param('I');
my $email  = CGI::param('E');
my $terms  = CGI::param('S');

if ($list =~ /\//) {
 ($list, $year, $month, $daydir, $id) = split (/\//, $list);
 $action = 'READ';
}

# Month hash
my %m = ('1'=>'January', '2'=>'February', '3'=>'March', '4'=>'April',
	'5'=>'May', '6'=>'June', '7'=>'July', '8'=>'August', '9'=>'September',
	'10'=>'October', '11'=>'November', '12'=>'December');

if ($list) {
 if (-s "$directory/$list/config") { do_conf("$directory/$list/config"); }
}

$template ||= "$directory/$list/template.html";
$template = "$directory/template.html" unless (-s $template);

my $page = '';
my $pagetemplate = '';
my $title = '';
my $showpage = 1;

if (-s $template)
{
    open TEMPLATE, "<$template";
    while (<TEMPLATE>) { $pagetemplate .= $_; }
    close TEMPLATE;
}
else
{
    $pagetemplate = qq(
	       <HTML><HEAD><TITLE><!-- MINORWEB TITLE --></TITLE>
	       <BODY BGCOLOR="#FFFFFF" TEXT="#000000">
	       <!-- MINORWEB BODY -->
	       </BODY></HTML>
	       );
}
$pagetemplate =~ s#<!-- *MINORWEB *HIDE -->.*?<!-- *MINORWEB */HIDE *-->##igs;

if ($action eq 'LIST')
{
    if (!($list)){
	my $lists = '';
	my $star = 'no';
	$title = "Choose a list";

	foreach (@main::archive)
	{
	    if ($_ eq "*")
	    {
		$lists = '';
		$star = 'yes';
		opendir(DIR, $directory);
		foreach (readdir(DIR)){ $lists.="$_ " if (-x "$directory/$_");}
		closedir(DIR);
	    }
	    elsif($star eq 'no')
	    {
		$lists .= "$_ ";
	    }
	}
	if ($lists ne '')
	{
	    $lists =~ s/\.\. //g;
	    $lists =~ s/\. //g;
	    @archive = split (/ /, $lists);
	}
        $page .= qq(<TABLE WIDTH="100%" BORDER="0">);
	foreach (sort(@archive)) {
	    $page.=qq(<TR><TD><A HREF="$minor_url?A=LIST&L=$_">$_</A></TD><TD>);
	    if (-r "$directory/$_/one-liner") {
		my $info = '';
		open INFO, "$directory/$_/one-liner";
		while (<INFO>) { $info .= $_ }
		close INFO;
		$page .= &html_quote($info);
	    }
	    $page .= "</TD></TR>";
	}
        $page .= "</TABLE>";

    } elsif ($year eq 'NONE'){
	opendir(DIR, "$directory/$list/archive");
	my @years = readdir(DIR);
	closedir(DIR);
	my $yearlist = join(" ", @years);
	$yearlist =~ s/\.\. //g;
	$yearlist =~ s/\. //g;
	@years = split(/ /, $yearlist);
	$title = "$list\@$main::domain";

	$page .= "<BLOCKQUOTE>";
        my $info = '';
	open INFO, "$directory/$list/info";
	while (<INFO>) { $info .= $_; }
	close INFO;
        $page .= &html_quote ($info);
	$page .= "</BLOCKQUOTE>";
	foreach $year (@years)
	{
	    $page .= "$year:";
	    opendir(DIR, "$directory/$list/archive/$year");
	    my @months = readdir(DIR);
	    closedir(DIR);
	    my $monthlist = join(" ", @months);
	    $monthlist =~ s/\.\. //g;
	    $monthlist =~ s/\. //g;
	    @months = split(/ /, $monthlist);
	    foreach (@months){
		my $m = substr ($m{$_}, 0, 3);
		$page .= qq( <A HREF="$minor_url?L=$list&Y=$year&M=$_">$m</A>);
	    }
	    $page .= "<BR>";
	}

    } else {
	my $sort = CGI::param('S') ? CGI::param('S') : 'DATE';
	my %days;
	my $day;
	my @subjs;
	my @names;
	my @emails;
	my @dates;
	my @di;
	$title = "$list\@$main::domain: $m{$month} $year";

	opendir(MONTH, "$directory/$list/archive/$year/$month");
	my @days = readdir(MONTH);
	closedir(MONTH);
	my $i = 0;
	foreach $day (@days)
	{
	    if ($day ne "." && $day ne "..")
	    {
		opendir (DAY, "$directory/$list/archive/$year/$month/$day");
		my @daylist = readdir(DAY);
		foreach $id (@daylist)
		{
		    unless ($id eq "." or $id eq "..")
		    {
			my $buf = '';
			open FILE, "$directory/$list/archive/$year/$month/$day/$id"; 
			while (<FILE>) { $buf .= $_ }
			close FILE;
			$buf =~ s/\r//g;
			my ($header, $message) = split(/\n\n/, $buf, 2);
			my @header = split(/\n/, $header);
			foreach (@header)
			{
			    if ($_ =~ /^from:/ig) { $names[$i] = $_; $emails[$i] = $_; }
			    elsif ($_ =~ /^date:/ig) { $dates[$i] = $_; }
			    elsif ($_ =~ /^subject:/ig) { $subjs[$i] = $_; }
			}
			$subjs[$i]  =~ s/^subject: //ig;
			$dates[$i]  =~ s/^date: //ig;
			$names[$i]  =~ s/^from: //ig;
			$names[$i] =~ s/<.*>//;
			$names[$i] =~ s/\"//g;
			$names[$i] =~ s/^.*\((.*)\).*$/$1/ if $names[$i] =~ /\(.*\)/;
			$emails[$i] =~ s/^from: //ig;
			$emails[$i] =~ s/.*\<//;
			$emails[$i] =~ s/\>.*//;
			$emails[$i] =~ s/\(.*\)//;
			$emails[$i] =~ s/\s//g;
			$di[$i]     = "$day:$id";
			if (!$subjs[$i]) { $subjs[$i] = "Unknown Subject" }
			$i++;
		    }
		}
		closedir(DAY);
	    }
	}
	
	$page .= qq(
		 <TABLE WIDTH="100%" BORDER="0">
		 <TR>
		 <TH ALIGN="CENTER">
		 <A HREF="$minor_url?L=$list/$year/$month&S=SUBJ">Subject</A>
		 </TH>
		 <TH ALIGN="CENTER">
		 <A HREF="$minor_url?L=$list/$year/$month&S=AUTH">Author</A>
		 </TH>
		 <TH ALIGN="CENTER">
		 <A HREF="$minor_url?L=$list/$year/$month&S=DATE">Date</A>
		 </TH>
		 </TR>
		 );
	
	# Sort the messages
	my @order = (0 .. $#subjs);
	
	if ($sort eq 'SUBJ') {
	    my %foo;
	    my @bar = ();
	    for $i (0 .. $#subjs) {
		my $baz = lc $subjs[$i];
		$baz =~ s/^ *//g;
		# Next two lines make original message come before Re:
		$baz =~ s/^re: *(.*)$/$1 MINOR:$i/ig;
		$baz .= " 0" if ($baz !~ / MINOR:$i$/);
		$foo{"$baz$i"} = $i;
	    }
	    @bar = sort keys %foo;
	    for $i (0 .. $#bar) { $order[$i] = $foo{$bar[$i]}; }
	}
	elsif ($sort eq 'AUTH') {
	    my %foo;
	    my @bar = ();
	    for $i (0 .. $#names) { my $baz = lc $names[$i]; $foo{"$baz$i"} = $i; }
	    @bar = sort keys %foo;
	    for $i (0 .. $#bar) { $order[$i] = $foo{$bar[$i]}; }
	}
	else {
	    # do nothing right now
	}
	
	foreach $i (@order) {
	    my ($d, $id) = split(/:/, $di[$i]);
	    $page .= qq(
			<TR>
			<TD ALIGN="LEFT">
			<A HREF="$minor_url?A=READ&L=$list/$year/$month/$d/$id">$subjs[$i]</A>
			</TD>
			<TD ALIGN="LEFT">
			<A HREF="mailto:$emails[$i]">$names[$i]</A>
			</TD>
			<TD ALIGN="LEFT">
			$dates[$i]
			</TD>
			</TR>
			);
	}
	
	$page .= "</TABLE>";
	
    }
    
} elsif ($action eq 'READ') {
    if (!($list && $year && $month)){
	print "Location: $minor_url?A=LIST&L=$list&Y=$year&M=$month\n\n";
	$showpage = 0;
    } else {
	# open up the header and read it in
	my $buf = '';
	open FILE, "$directory/$list/archive/$year/$month/$daydir/$id";
	while (<FILE>) { $buf .= $_ }
	close FILE;
	$buf = &html_quote ($buf);
	my ($header, $message) = split(/\n\n/, $buf, 2);
	my @header = split(/\n/, $header);
	my $from = '';
	my $date = '';
	my $to = '';
	my $cc = '';
	my $subject = '';
	my $boundary = '';
	my $last = 'last';
	foreach (@header){
	    if ($_ =~ /^\s/) {
		$from .= $_ if $last eq "from";
		$date .= $_ if $last eq "date";
		$to .= $_ if $last eq "to";
		$cc .= $_ if $last eq "cc";
		$subject .= $_ if $last eq "subject";
		$boundary .= $_ if $last eq "boundary";
	    }
	    elsif ($_ =~ /^from:/ig) { $from = $_; $last = "from"; }
	    elsif ($_ =~ /^date:/ig) { $date = $_; $last = "date"; }
	    elsif ($_ =~ /^to:/ig) { $to = $_; $last = "to"; }
	    elsif ($_ =~ /^cc:/ig) { $cc = $_; $last = "cc"; }
	    elsif ($_ =~ /^subject:/ig) { $subject = $_; $last = "subject"; }
	    elsif ($_ =~ /^content-type: multipart/ig) {
		$boundary = $_;
		$last = "boundary";
	    }
	    else { $last = 'last'; }
	}
	
	$title = $subject;
	$title =~ s/^subject: //i;
	$title = "$list\@$main::domain: $title";
	
	my @attlist = ();
	
	# MIME attachment handling goes here
	if ($boundary) {
	    $boundary =~ s/^.*\"(.*)\".*$/$1/gs;
	    $message =~ s/($boundary)\-.*/$1/g;
	    my @parts = split(/$boundary/, $message);
	    my $att = CGI::param('ATT') ? CGI::param('ATT') : 0;
	    if ($att == 0) {
		my $j = 0;
		my $i = 0;
		$message = '';
		foreach (@parts) {
		    my ($h, $b) = split(/\n\r*\n/, $_, 2);
		    if ($h =~ /text\/plain/igs || $h !~ /content-type/igs) {
			$message .= $b;
		    }
		    else {
			$h =~ s/^.*name=\"(.*)\".*$/$1/igs if $h =~ /name=/igs;
			$attlist[$j++] = "$i:$h";
		    }
		    $i++;
		}
	    }
	    else
	    {
		my ($header, $body) = split(/\n\r*\n/, $parts[$att], 2);
		#$header =~ s/Content-Transfer-Encoding.*\n//ig;
		$body =~ s/--$//gs;
		print $header;
		print "\n\n";
		print decode_base64($body);
		$showpage = 0;
	    }
	}
	
	if ($showpage)
	{
	    $page .= "<PRE>";
	    $page .= "$date\n" if $date;
	    $page .= "$from\n" if $from;
	    $page .= "$to\n" if $to;
	    $page .= "$cc\n" if $cc;
	    $page .= "$subject\n" if $subject;
	    $page .= "\n</PRE>";
	    $page .= "<PRE>$message</PRE>";
	    foreach (@attlist) {
		my ($num, $name) = split(/:/, $_, 2);
		$name = "Unknown" if !$name;
		$page .= "<A HREF=\"$minor_url?A=READ&L=$list&Y=$year&M=$month&D=$daydir&I=$id&ATT=$num\">Attachment: $name</A><BR>";
		#$page .= "<IMG SRC=\"$minor_url?A=READ&L=$list&Y=$year&M=$month&D=$daydir&I=$id&ATT=$num\" ALT=\"$name\"><BR>" if $name =~ /.gif$/i;
	    }
	}
    }
} elsif ($action eq 'SUB') {
    # Subscribe
    &sub_unsub ($email, 'subscribe', $list);
    $title = "Subscription request sent";
    $page .= qq(
		Thanks for subscribing to $list\@$main::domain. You should
		recieve	a confirmation e-mail shortly.
		);
} elsif ($action eq 'UNSUB') {
    # Unsubscribe
    &sub_unsub ($email, 'unsubscribe', $list);
    $title = "Unsubscription request sent";
    $page .= qq(
		You should receive a confirmation e-mail shortly regarding
		your request to unsubscribe from $list\@$main::domain.
		);
} elsif ($action eq 'SEARCH') {
#FIXME: search!
	$title = "Search $list for $terms";
	$page .= "Sorry, search isn't implemented yet";
}

if ($showpage)
{
    if ($list)
    {
        # Subscribe/unsubscribe form
        $page .= qq(
		    <P ALIGN="CENTER">
		    <FORM METHOD="GET" ACTION="$minor_url">
		    E-mail address: <INPUT TYPE="TEXT" NAME="E" SIZE="24">
		    <SELECT NAME="A" SIZE="1">
		    <OPTION VALUE="SUB">Subscribe</OPTION>
		    <OPTION VALUE="UNSUB">Unsubscribe</OPTION>
		    </SELECT>
		    <INPUT TYPE="HIDDEN" NAME="L" VALUE="$list">
		    <INPUT TYPE="SUBMIT" VALUE="Go">
		    </FORM>
		    </P>
		    );
        if (1 == 0) {
	# FIXME: search is unimplemented!
	# Search form
	$page .= qq(
		    <P ALIGN="CENTER">
		    <FORM METHOD="GET" ACTION="$minor_url">
		    Search for: <INPUT TYPE="TEXT" NAME="S" SIZE="24">
		    <INPUT TYPE="HIDDEN" NAME="A" VALUE="SEARCH">
		    <INPUT TYPE="HIDDEN" NAME="L" VALUE="$list">
		    <INPUT TYPE="SUBMIT" VALUE="Search">
		    );

	$page .= "<BR>Between ";
        opendir(DIR, "$directory/$list/archive");
        my @years = readdir(DIR);
        closedir(DIR);
        my $yearlist = join(" ", @years);
        $yearlist =~ s/\.\. //g;
        $yearlist =~ s/\. //g;
        @years = sort(split(/ /, $yearlist));

        $page .= qq(<SELECT NAME="FD" SIZE="1">);
	foreach (1 .. 31)
	{
	    #FIXME: first day
	    my $sel = '';
	    $page .= qq(<OPTION VALUE="$_" $sel>$_</OPTION>);
	}
	$page .= "</SELECT>";

        $page .= qq(<SELECT NAME="FM" SIZE="1">);
	foreach (1 .. 12)
	{
	    #FIXME: first month
	    my $sel = '';
	    $page .= qq(<OPTION VALUE="$_" $sel>$m{$_}</OPTION>);
	}
	$page .= "</SELECT>";

	$page .= qq(<SELECT NAME="FY" SIZE="1">);
	foreach (@years)
	{
	   my $sel = ($years[0] == $_) ? "SELECTED" : '';
	   $page .= qq(<OPTION VALUE="$_" $sel>$_</OPTION>);
	}
	$page .= "</SELECT>";

	$page .= "<BR>and ";

        $page .= qq(<SELECT NAME="LD" SIZE="1">);
	foreach (1 .. 31)
	{
	    #FIXME: last day
	    my $sel = '';
	    $page .= qq(<OPTION VALUE="$_" $sel>$_</OPTION>);
	}
	$page .= "</SELECT>";

        $page .= qq(<SELECT NAME="LM" SIZE="1">);
	foreach (1 .. 12)
	{
	    #FIXME: last month
	    my $sel = '';
	    $page .= qq(<OPTION VALUE="$_" $sel>$m{$_}</OPTION>);
	}
	$page .= "</SELECT>";

	$page .= qq(<SELECT NAME="LY" SIZE="1">);
	foreach (@years)
	{
	   my $sel = ($years[-1] == $_) ? "SELECTED" : '';
	   $page .= qq(<OPTION VALUE="$_" $sel>$_</OPTION>);
	}
	$page .= "</SELECT>";

	$page .= "</FORM>";
        }
        $page .= "</P>";
    }
    $pagetemplate =~ s/<!-- *MINORWEB *TITLE *-->/$title/igs;
    $pagetemplate =~ s/<!-- *MINORWEB *BODY *-->/$page/igs;
    $pagetemplate =~ s/<!--.*? -->//igs;
    print "Content-type: text/html\n\n";
    print $pagetemplate;
}

###########################################################################
# Reads a configuration file
###########################################################################
sub do_conf
{
    # Get the file to read and initialize variables
    my $file = shift;
    my $directive = '';
    my $archive = '';
    my $list_cmd = '';
    my $open_lists = '';
    
    # Parse the thing
    open CONF, $file;
    while (<CONF>){
	if($_ =~ /^#/){
	   # Comments start with a #
       } elsif ($_ =~ /^smtp_server/i){
	   ($directive, $main::smtp_server) = split(/=/, $_, 2);
	   chomp $main::smtp_server;
       } elsif ($_ =~ /^directory/i){
	   ($directive, $main::directory) = split(/=/, $_, 2);
	   chomp $main::directory;
       } elsif ($_ =~ /^admin/i){
	   ($directive, $main::admin) = split(/=/, $_, 2);
	   chomp $main::admin;
       } elsif ($_ =~ /^domain/i){
	   ($directive, $main::domain) = split(/=/, $_, 2);
	   chomp $main::domain;
       } elsif ($_ =~ /^archive/i){
	   ($directive, $main::archive) = split(/=/, $_, 2);
	   chomp $main::archive;
       } elsif ($_ =~ /^list_cmd/i){
	   ($directive, $main::list_cmd) = split(/=/, $_, 2);
	   chomp $main::list_cmd;
       } elsif ($_ =~ /^open_lists/i){
	   ($directive, $main::open_lists) = split(/=/, $_, 2);
	   chomp $main::open_lists;
       } elsif ($_ =~ /^subject_prefix/i){
	   ($directive,$main::subject_pre)=split(/=/, $_, 2);
	   chomp $main::subject_pre;
       } elsif ($_ =~ /^url/i){
	   ($directive, $main::url) = split(/=/, $_, 2);
	   chomp $main::url;
       } elsif ($_ =~ /^template/i){
	   ($directive, $main::template) = split (/=/, $_, 2);
	   chomp $main::template;
       }
	
    }
    close CONF;
    
    # These are space separated
    @main::archive = split(/ /, $main::archive);
    @main::list_cmd = split(/ /, $main::list_cmd);
    @main::open_lists = split(/ /, $main::open_lists);
    
    # These should be recognized as regular expressions
    $main::subject_pre =~ s/[\r|\n]//g;
    $main::subject_pre =~ s/([\[|\]])/\\$1/g;
}

###########################################################################
# Send a message
###########################################################################
sub sub_unsub {
 # Open the socket
 socket (SOCK, PF_INET, SOCK_STREAM, getprotobyname ('tcp'));
 connect (SOCK, sockaddr_in (25, inet_aton ($main::smtp_server)));
 
 # Get the variables
 my $from = shift;
 my $action = shift;
 my $list = shift;
 my $err = '';

 my $message = "From: $from\r\n";
 $message .= "To: minordomo\@$main::domain\r\n";
 $message .= "Subject: $action $list\r\n";
 $message .= "Date: ";
 $message .= &POSIX::strftime ("%a, %d %b %Y %T %z", localtime);
 $message .= "\r\n\r\n";

 # SMTP session
 send SOCK, "HELO $main::domain\r\n", 0;
 recv SOCK, $err, 512, 0;
 send SOCK, "MAIL FROM:<$from>\r\n", 0;
 recv SOCK, $err, 512, 0;
 send SOCK, "RCPT TO:<minordomo\@$main::domain>\r\n", 0;
 recv SOCK, $err, 512, 0;
 send SOCK, "DATA\r\n", 0;
 recv SOCK, $err, 512, 0;
 send SOCK, "$message\r\n.\r\n", 0;
 recv SOCK, $err, 512, 0;
 send SOCK, "QUIT\r\n", 0;
 recv SOCK, $err, 512, 0;
 close SOCK;
} 

sub html_quote
{
  my $buf = shift;
  $buf =~ s/</&lt\;/g;
  $buf =~ s/>/&gt\;/g;
  $buf =~ s/\r//g;
  foreach (qw(http ftp nntp news gopher telnet)) {
    $buf =~ s/($_:[\d|\w|\/|\.|\-|?|&|=|~]*[\d|\w|\/])([\b|\n| ]*)/<A HREF=\"$1\">$1<\/A>$2/gs;
  }
  $buf =~ s/(\w[\w|\d|\.|\-|\+]*\@[\w|\d|\.|\-]*)([\b|\n| ]*)/<A HREF=\"mailto:$1\">$1<\/A>$2/gs;
  $buf =~ s/(&lt\;)(\w[\w|\d|\.|\-|\+]*\@[\w|\d|\.|\-]*)(&gt\;)/$1<A HREF=\"mailto:$2\">$2<\/A>$3/gs;
  return $buf;
}

###########################################################################
# decode_base64 - decodes BASE64 encoded MIME attachments
# Used under terms of the GPL from the MIME::Base64 module
###########################################################################
sub decode_base64 ($)
{
    local($^W) = 0; # unpack("u",...) gives bogus warning in 5.00[123]

    my $str = shift;
    my $res = "";

    $str =~ tr|A-Za-z0-9+=/||cd;            # remove non-base64 chars
    $str =~ s/=+$//;                        # remove padding
    $str =~ tr|A-Za-z0-9+/| -_|;            # convert to uuencoded format
    while ($str =~ /(.{1,60})/gs) {
        my $len = chr(32 + length($1)*3/4); # compute length byte
        $res .= unpack("u", $len . $1 );    # uudecode
    }
    $res;
}
